<?php

/**
 * Functions for calculating Latent Semantic Analysis/Singular Vector Decomposition 
 * for idea charactiersitcs matrix (characteristics x ideas )
 */
require_once "JAMA/Matrix.php";

/**
 * This function creates the form to update RDF namespaces
 */
function idea_analytics_cluster_analysis($form_state, $page_number = 0) {

  /******get all available relationship types******/
  $clustering_types = array('','Tags', 'Categories');
  //contitional clustering type: check if Search module is available (Frequent Keywords are taken from tokens extracted by the Drupal indexing engine)
  if (module_exists('search')) //if module is enabled 
  	$clustering_types[] = 'Frequent keywords';
  
  //contitional clustering type: check if Gi2MO Types categorisation is installed to add clustering by Idea Characteristics
  
  //get 'Idea Classification' vocabulary id
  $result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Idea Classification'");
  if ($data = db_fetch_object($result)) {
   $vocabulary_id=$data->vid;
  }
	
  if (isset($vocabulary_id) && user_access('view idea characteristics')) //if idea classification vocabulary exists and user has access to view it
  	$clustering_types[] = 'Characteristics';

  //contitional clustering type: check if IdeaStream Similarity is available and user has access to view to add clustering by Relationships
  if (module_exists('similar_ideas') && user_access('view similar ideas')) //if module is enabled and user has access to view relationships
  	$clustering_types[] = 'Relationships';
  	
  	
  /********* get clustering settings *********/
  $advanced_settings_default = array(
  	'Tags' => array (
  				'cluster_count' => 4,
  				'max_terms' => 50,
  				'blacklist' => ''),
  	'Categories' => array (
  				'cluster_count' => 4,
  				'max_terms' => 50,
  				'blacklist' => ''),
  	'Frequent keywords' => array (
  				'cluster_count' => 4,
  				'max_terms' => 50,
  				'blacklist' => "the,and,for,that,not,you,add,idea,make,more,all,there,have,is,this,with,should,others,nothing,are,but,can,like,would,use,when,from,could,better,some,they,ideas,what,just,also,will,think,easy,which,need,1"),
  	'Characteristics' => array (
  				'cluster_count' => 4,
  				'max_terms' => 90,
  				'blacklist' => ''),
  	'Relationships' => array (
  				'cluster_count' => 5,
  				'max_terms' => 60,
  				'blacklist' => ''),
  );
   
  $advanced_settings = variable_get('ideastream_analytics_advanced_settings',$advanced_settings_default);

  //check if form was resubmitted and save the new advanced settings if needed
  if (isset($form_state['values']["advanced_settings"]) && isset($form_state["values"]['chart_cluster_criteria']))
  {
  	if ($form_state["values"]['chart_cluster_criteria']!=0) {
  	  if (isset($form_state['values']["advanced_settings"]["cluster_count"]))
  	  	if (is_numeric($form_state['values']["advanced_settings"]["cluster_count"]))
	  		$advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['cluster_count'] = $form_state['values']["advanced_settings"]["cluster_count"];
	  	else drupal_set_message("Advanced Settings Error: <b>Max. cluster count</b> has to be an integer!",'error');
	  	
  	  if (isset($form_state['values']["advanced_settings"]["term_list_length"]))
  	  	if (is_numeric($form_state['values']["advanced_settings"]["term_list_length"]))
	  		$advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['max_terms'] = $form_state['values']["advanced_settings"]["term_list_length"];
	  	else drupal_set_message("Advanced Settings Error: <b>Max. term count</b> has to be an integer!",'error');

  	  if (isset($form_state['values']["advanced_settings"]["term_blacklist"]))
	  		$advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['blacklist'] = $form_state['values']["advanced_settings"]["term_blacklist"];
	  variable_set('ideastream_analytics_advanced_settings',$advanced_settings);
	  } else drupal_set_message("Please select clustering criteria type. The criteria indicates which annotations are analysed in order to group ideas together.",'warning');
  } 

  //add libraries used for popups  
  drupal_add_js(drupal_get_path('module', 'ideastream_analytics') . '/js/jquery-1.8.3.min.js');
  drupal_add_js(drupal_get_path('module', 'ideastream_analytics') . '/js/jquery.tipsy.js');
  drupal_add_css(drupal_get_path('module', 'ideastream_analytics') .'/js/tipsy.css');

  //add the main javascript for drawing the plot with d3 library
  drupal_add_js(drupal_get_path('module', 'ideastream_analytics') . '/js/ideastream_analytics_cluster_plot.js');


	$js = drupal_add_js();										
	
	//add advanced settings so that they are visible by javascript
  	drupal_add_js(array('ideastream_analytics' => array(
													'advanced_settings' => $advanced_settings)), 'setting');	
  	
	//display form elements  
    $form['cluster_statistics'] = array(
    '#type' => 'fieldset',
    '#title' => t('Cluster Analysis Overview'),
    '#collapsible' => FALSE,
    '#collapsed' => FALSE,
    '#tree' => FALSE,
    '#description' => t('Cluster analysis compares ideas and attempts to organise them in groups based on particular criteria of analysis. <br/><br/> Calculations can be made either based on semantic analysis of natural language idea description or based on analysis idea annotations (like tags, categories etc.)'),
  	);
  
    $form['cluster_statistics']['chart_cluster_criteria'] = array(
    '#type' => 'select',
    '#title' => t('Cluster Criteria'), 
    '#options' => $clustering_types,
    '#default_value' => $form_state["values"]['chart_cluster_criteria'],
    '#prefix' => '<div class="container-inline">',
  );
    
  $form['cluster_statistics']['calculate'] = array(
    '#type' => 'submit',
    '#value' => 'Calculate',
    '#suffix' => '</div>',
  );
  
  $form['cluster_statistics']['advanced_settings'] = array(
    '#type' => 'fieldset',
    '#title' => t('Advanced Settings'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    '#tree' => TRUE,
    '#description' => t('<p>The clustering algorithm can have different output depending on variety of parameters:<br/>
    <b>*Max. cluster count</b> - max. number of groups that algorithm should aim to find<br/>
    <b>*Max. term count</b> - amount of unique terms (e.g. tags, categories) used for detecting clusters. The more terms, the more time the clustering algorithm takes. This setting allows to define the amount of top used terms (most frequent) to be considered in the clustering process (e.g. top 50 used tags).<br/>
    <b>*Term blacklist</b> - terms ignored during cluster analysis. Sometimes the top used terms or common words (\'is\',  \'are\' etc.) are so frequent that they do not add anything to understanding of groups (e.g. all ideas have a term).<br/>
    </p>'),
  );
  
  $form['cluster_statistics']['advanced_settings']['cluster_count'] = array (
    '#type' => 'textfield',
    '#title' => t('Max. cluster count'), 
    '#size' => 5,
    '#default_value' => $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['cluster_count'],  //by default 4 clusters
    '#maxlength' => 10,
    '#prefix' => '<div class="container-inline">',
  );
    
  $form['cluster_statistics']['advanced_settings']['term_list_length'] = array (
    '#type' => 'textfield',
    '#title' => t('Max. term count'), 
    '#size' => 5,
    '#default_value' => $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['max_terms'],  //by default 4 clusters
    '#maxlength' => 10,
	'#suffix' => '</div>',
  );
  
  $form['cluster_statistics']['advanced_settings']['term_blacklist'] = array (
    '#type' => 'textfield',
    '#title' => t('Term blacklist'), 
    '#size' => 90,
    '#default_value' => $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['blacklist'],  //by default 4 clusters
    '#maxlength' => 99999999,
  );
  
    $form['chart'] = array(
    '#type' => 'fieldset',
    '#title' => t('Cluster Analysis Visualisation'),
    '#collapsible' => FALSE,
    '#collapsed' => FALSE,
    '#tree' => FALSE,
    '#description' => t('The visualisation shows groups of ideas and semantic interpretation of those groups related to selected criteria.')
  	);
  	
  	$blacklist = $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['blacklist'];
  	//check for options and if clustering type has been chosen: calculate clusters and render the chart
  	if ($form_state["values"]['chart_cluster_criteria']>0)
  	{
  		//add d3 javascript to the page
  		drupal_add_js(drupal_get_path('module', 'ideastream_analytics') . '/js/d3.v3.min.js');
  		
  		//add styles for the chart
		drupal_add_css(drupal_get_path('module', 'ideastream_analytics') . '/ideastream_analytics.css');

		//prepare settings array
		$settings_selected = array();
		$settings_selected['cluster_count'] = $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['cluster_count'];
		$settings_selected['max_terms'] = $advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['max_terms'];
		$settings_selected['blacklist'] = array_map('trim',explode(',',$advanced_settings[$clustering_types[$form_state["values"]['chart_cluster_criteria']]]['blacklist']));
		//process idea data for clustering
		$cluster_criteria_type = $clustering_types[$form_state["values"]['chart_cluster_criteria']]; //categories, tags, characteristics etc.
  		$clustered_data = idea_analytics_cluster_analysis_calculate($cluster_criteria_type,$settings_selected);
  		
  		if (is_array($clustered_data)) {
  		
	  		//post-process cluster data to format it properly for the d3 javascript library
	  		$cluster_centers = $clustered_data['centroids'];
	  		unset($clustered_data['centroids']);
	  		
	  		$cluster_data_joint = array(); //array with all clusters joint together and cluster names put as additional value in each row (used later for d3 visualisation)
	  		$cluster_number = 0;
	  		//var_dump($clustered_data);
	  		
	  		//arrays used to display data statistics below the chart
	  		$ideas_clusters = array(); //array with ideas per cluster
	  		$tags_clusters = array(); //array with tags/characteristics/keywords per cluster
	  		
	  		$cluster_data_tags_temp = array();
	  		
	  		
	  		foreach ($clustered_data as $cluster)
	  		{
	  			$cluster_number++;
	  			foreach ($cluster as $point_coords)
	  			{
	  				if ($point_coords[2][0] != 'idea') //check if its a regular idea data point of tag/characteristic/keyword coord node
	  				{
	  					$point_coords[] = $point_coords[2][0]; //add as 'cluster name' the type of data that was used for clustering (characteristics, tags etc.) - this creates a separate cluster for visualisation with characteristics/keywords/tags etc.
	  					$tags_clusters[$cluster_number-1][] = $point_coords; //add to separate array used for statistics
	  					$cluster_data_tags_temp[] = $point_coords; // temp array that is added to the end of joint one (to keep all idea coords before tag coords)
	  				}
	  				else
	  				{
	  					$point_coords[] = "Cluster " . $cluster_number; //if regular data add to cluster
	  					$ideas_clusters[$cluster_number-1][] = $point_coords; //add to separate array used for statistics
	  					$cluster_data_joint[] = $point_coords;
	  				}
	  			}	
	  		}
	  		
	  		$cluster_data_joint = array_merge($cluster_data_joint,$cluster_data_tags_temp); //join tags coords to the main array (so that tags are displayed in chart too as last cluster)
	  		//pass clustered information to javascript
	  		drupal_add_js(array('ideastream_analytics' => array(
														'cluster_data' => $cluster_data_joint)
														), 'setting');
																											
	  		//display cluster chart
		  	$form['chart']['custom_html'] = array(
		      '#value' => '<div id="ideaanalytics_cluster_plot"></div>',
		  	);
		  	
		  	//display cluster statistics
		  	$statistics_html = idea_analytics_cluster_analysis_clusters_stats_overview($ideas_clusters, $tags_clusters, $cluster_criteria_type, $settings_selected['blacklist']);
		  	$form['chart']['custom_statistics_html'] = array(
		      '#value' => $statistics_html,
		  	);
	  	}
	  	else  //if clustering failed
	  	{
	  		if ($clustered_data==-1)
	  			drupal_set_message(" Error (-1): No terms detected. There are no ideas annotated with terms of selected type!",'error');
	  	}
  	}
  
  return $form;
}

/*
 * Takes as input ideas and tags with ids and names split into clusters and outs HTML that presents an overview of statistics of clusters
 * @param ideas_clusters[cluster_number][idea_data] - two dimensional array with ideas per cluster. First  dimension is cluster, second is ideas where each row is [coordX, coordY, ['idea', idea_nid, idea_title], cluster_name]]
 * @param tags_clusters[cluster_number][criteria_data] - array with information about clustering criterias, structure similar to ideas_clusters
 * @param cluster_criteria_type a string the identified criteria for clustering used (Tags, Categories, Characteristics etc.)
 * @return html with the overview statistics of the clusters
 */
function idea_analytics_cluster_analysis_clusters_stats_overview($ideas_clusters, $tags_clusters, $cluster_criteria_type, $blacklist) {
	
	$d3_color_scale10 = array ('#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf');
	
	$cluster_count = count($ideas_clusters);
	
	$page_html="";
	
	/************ CALCULATE STATISTICS PER EACH CLUSTER **********/
	$idea_ids_per_cluster = array();
	$idea_count_per_cluster = array();
	$comment_count_per_cluster = array();
	$comment_idea_ratio_per_cluster = array();
	$term_count_per_cluster = array();
	$term_ids_per_cluster = array();
	$workflow_idea_count_per_cluster = array();
	
	$current_cluster_number = 0;
	foreach ($ideas_clusters as $idea_cluster)
	{
		//term count per cluster
		$tag_cluster = $tags_clusters[$current_cluster_number];
		$tag_cluster_count = count($tag_cluster);
		$term_count_per_cluster[$current_cluster_number] = $tag_cluster_count;
		
		//term ids per cluster
		if ($tag_cluster)
			foreach ($tag_cluster as $tag) 
				$term_ids_per_cluster[$current_cluster_number][$tag[2][2]] = $tag[2][1];
		
		//idea count per cluster
		$idea_count = count($idea_cluster);
		$idea_count_per_cluster[] = $idea_count; //save for later use in overall stats

		//get cumulated comment count for all ideas in the cluster

		$idea_ids = array();
		foreach ($idea_cluster as $idea_info)
			$idea_ids[] = $idea_info[2][1];
		
		$idea_ids_text = implode(",",$idea_ids);
		
		//add to global table for later processing
		$idea_ids_per_cluster[] = $idea_ids;
		
	  	$result = db_query("SELECT SUM(comment_count) AS total_comment_count FROM node_comment_statistics WHERE nid IN ($idea_ids_text)");
		
		//comment count stat
		$commnet_count = 0;
		if ($data = db_fetch_object($result))
			$commnet_count=$data->total_comment_count;
		
		$comment_count_per_cluster[] = $commnet_count;
		
		//comment count per idea count ratio
		$comment_idea_ratio_per_cluster[] = round($commnet_count / $idea_count,2);
		
		//idea status stats
		$idea_statues_stats = idea_analytics_metric_ideas_per_status_count($idea_ids);
		foreach ($idea_statues_stats as $idea_state_stat)			
			$workflow_idea_count_per_cluster[$idea_state_stat->STATE_NAME][$current_cluster_number]=$idea_state_stat->IDEA_COUNT;
			
		$current_cluster_number++;
	}
		
	// calculate min/max values of metrics out of all clusters
	$max_idea_count = max($idea_count_per_cluster);
	$min_idea_count = min($idea_count_per_cluster);
	
	$max_comment_count = max($comment_count_per_cluster);
	$min_comment_count = min($comment_count_per_cluster);

	$max_comment_idea_ratio = max($comment_idea_ratio_per_cluster);
	$min_comment_idea_ratio = min($comment_idea_ratio_per_cluster);
	
	foreach ($workflow_idea_count_per_cluster as $workflow_type_name => $workflow_type_idea_counts)
	{
		$max_idea_count_per_workflow_state[$workflow_type_name] = max($workflow_type_idea_counts);
		$min_idea_count_per_workflow_state[$workflow_type_name] = min($workflow_type_idea_counts);
	}
	
	//calculate idea count per each term in every cluster (this is extracted from database and query varies depending on criteria type)
	$current_cluster_number = 0;
	$idea_count_per_term_per_cluster = array();
	foreach ($ideas_clusters as $idea_cluster)
	{
		if (isset($idea_ids_per_cluster[$current_cluster_number]))
			$idea_ids_text = implode(',',$idea_ids_per_cluster[$current_cluster_number]);
		else
			$idea_ids_text = 'NULL'; 
		switch($cluster_criteria_type) {
			case 'Characteristics':
				if (isset($term_ids_per_cluster[$current_cluster_number]))
					$term_ids_text = implode(',',$term_ids_per_cluster[$current_cluster_number]);
				else 
					$term_ids_text = "NULL";
				
				$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Idea Classification'");
				if ($data = db_fetch_object($result)) {
				   $vocabulary_id=$data->vid;
				}
	
				if (isset($vocabulary_id)) //if idea classification vocabulary exists
				{
					if (isset($blacklist) && $blacklist!="")
						$blacklist_string = "'".implode("','", $blacklist)."'";
					else 
						$blacklist_string = "''";
					$query = "SELECT COUNT(node.nid) AS idea_count, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND term_data.vid=$vocabulary_id AND node.nid IN ($idea_ids_text) AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY idea_count DESC LIMIT 0,50"; //AND term_data.tid IN ($term_ids_text)
				}
				break;
			case 'Tags':
				if (isset($term_ids_per_cluster[$current_cluster_number]))
					$term_ids_text = implode(',',$term_ids_per_cluster[$current_cluster_number]);
				else 
					$term_ids_text = "NULL";
					
				$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Tags'");
				if ($data = db_fetch_object($result)) {
				   $vocabulary_id=$data->vid;
				}
	
				if (isset($vocabulary_id)) //if idea tags vocabulary exists
				{
					if (isset($blacklist) && $blacklist!="")
						$blacklist_string = "'".implode("','", $blacklist)."'";
					else 
						$blacklist_string = "''";
					$query = "SELECT COUNT(node.nid) AS idea_count, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND term_data.vid=$vocabulary_id AND node.nid IN ($idea_ids_text) AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY idea_count DESC LIMIT 0,50"; // AND term_data.tid IN ($term_ids_text)
				}
				break;
			case 'Categories':
				if (isset($term_ids_per_cluster[$current_cluster_number]))
					$term_ids_text = implode(',',$term_ids_per_cluster[$current_cluster_number]);
				else 
					$term_ids_text = "NULL";
				
				$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Categories'");
				if ($data = db_fetch_object($result)) {
				   $vocabulary_id=$data->vid;
				}
	
				if (isset($vocabulary_id)) //if idea categories vocabulary exists
				{
					if (isset($blacklist) && $blacklist!="")
						$blacklist_string = "'".implode("','", $blacklist)."'";
					else 
						$blacklist_string = "''";
					$query = "SELECT COUNT(node.nid) AS idea_count, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND term_data.vid=$vocabulary_id AND node.nid IN ($idea_ids_text) AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY idea_count DESC LIMIT 0,50"; // AND term_data.tid IN ($term_ids_text)
				}
				break;
			case 'Frequent keywords':
				if (isset($term_ids_per_cluster[$current_cluster_number]))
					$term_ids_text = "'".implode("','",$term_ids_per_cluster[$current_cluster_number])."'";
				else 
					$term_ids_text = "NULL";
					
					if (isset($blacklist) && $blacklist!="")
						$blacklist_string = "'".implode("','", $blacklist)."'";
					else 
						$blacklist_string = "''";
				$query = "SELECT COUNT(sid) AS idea_count, word AS term_name FROM search_index WHERE sid IN ($idea_ids_text) AND  word NOT IN ($blacklist_string) GROUP BY term_name ORDER BY idea_count DESC LIMIT 0,50"; //word IN ($term_ids_text) AND
				break;
			case 'Relationships':
					if ($term_ids_per_cluster[$current_cluster_number] == NULL) $term_ids_text = "''"; else
						$term_ids_text = "'".implode("','",$term_ids_per_cluster[$current_cluster_number])."'";
					if (isset($blacklist) && $blacklist!="")
						$blacklist_string = "'".implode("','", $blacklist)."'";
					else 
						$blacklist_string = "''";
					$query = "SELECT COUNT(idea1) AS idea_count, relationship_type AS term_name FROM similar_ideas_relations WHERE idea1 IN ($idea_ids_text) AND relationship_type NOT IN ($blacklist_string) GROUP BY term_name ORDER BY idea_count DESC LIMIT 0,50"; //relationship_type IN ($term_ids_text) AND
				break;
			default:		
				break;
		}
  		$result = db_query($query);
  		$special_term_per_cluster = array(); //list of terms by name closest to the cluster based on SVD decomposition
		while ($data = db_fetch_object($result))
			$idea_count_per_term_per_cluster[$current_cluster_number][$data->term_name] = $data->idea_count;		
		$current_cluster_number ++;
	}
	
	/****************** PRINT OVERALL SUMMARY STATISTICS FOR THE ENTIRE CLUSTERING OUTPUT **************/
	//print the overall statistics (they go before individual clusters but are generated afterwards because data tables are created in the loop above)
	$page_html_overall_stats .= '<table class="analytics_metrics_panel" border=0 style="table-layout: fixed;">';
	// ************ TIME STATS
	
	//calculate stats
	
	//calculate last idea date
	$idea_ids_all = array();
	foreach ($idea_ids_per_cluster as $idea_cluster_ids)
		$idea_ids_all = array_merge($idea_ids_all, $idea_cluster_ids);
	
	$idea_ids_all_text = implode(',',$idea_ids_all);
  	$result = db_query("SELECT node.nid AS IDEA_ID, FROM_UNIXTIME(node.created) AS IDEA_CREATED FROM node WHERE node.type= 'idea' AND node.status = 1 AND node.nid IN ($idea_ids_all_text) ORDER BY created DESC");
	
	//last idea date
	$last_idea_date = NULL;
	$last_idea_id = NULL;
	if ($data = db_fetch_object($result))
	{
		$last_idea_date=$data->IDEA_CREATED;	
		$last_idea_id=$data->IDEA_ID;
	}
	
	//idea count (avg, min, max)
	$idea_count_avg = NULL;
	$idea_count_max = NULL;
	$idea_count_min = NULL;
	$idea_count_total = NULL;
	foreach ($idea_count_per_cluster as $idea_count)
	{
		$idea_count_total += $idea_count;
		if ($idea_count_max==NULL || ($idea_count_max!=NULL && $idea_count > $idea_count_max)) $idea_count_max = $idea_count;
		if ($idea_count_min==NULL || ($idea_count_min!=NULL && $idea_count < $idea_count_min)) $idea_count_min = $idea_count;
	}
	if ($cluster_count!=NULL && $cluster_count!=0)
		$idea_count_avg = $idea_count_total / $cluster_count;
		
	//comment count (avg, min, max)
	$comment_count_avg = NULL;
	$comment_count_max = NULL;
	$comment_count_min = NULL;
	$comment_count_total = NULL;
	foreach ($comment_count_per_cluster as $comment_count)
	{
		$comment_count_total += $comment_count;
		if ($comment_count_max==NULL || ($comment_count_max!=NULL && $comment_count > $comment_count_max)) $comment_count_max = $comment_count;
		if ($comment_count_min==NULL || ($comment_count_min!=NULL && $comment_count < $comment_count_min)) $comment_count_min = $comment_count;
	}
	if ($cluster_count!=NULL && $cluster_count!=0)
		$comment_count_avg = $comment_count_total / $cluster_count;
	
	
	//**** save all stats to SESSION (so that later all can be also exported to CSV without repeating the calculations all over again)
  
	if (isset($_SESSION['cluster_analysis'])) //if the variable is set delete previous information
		unset($_SESSION['cluster_analysis']); //=$_SESSION['views']+1;
	
	$_SESSION['cluster_analysis']['overall_stats']['cluster_criteria']=$cluster_criteria_type;
	$_SESSION['cluster_analysis']['overall_stats']['cluster_count']=$cluster_count;
	$_SESSION['cluster_analysis']['overall_stats']['idea_count_total']=$idea_count_total;
	$_SESSION['cluster_analysis']['overall_stats']['last_idea_date']=$last_idea_date;
	
	$_SESSION['cluster_analysis']['overall_stats']['idea_count_avg']=$idea_count_avg;
	$_SESSION['cluster_analysis']['overall_stats']['idea_count_max']=$idea_count_max;
	$_SESSION['cluster_analysis']['overall_stats']['idea_count_min']=$idea_count_min;
	
	$_SESSION['cluster_analysis']['overall_stats']['comment_count_avg']=$comment_count_avg;
	$_SESSION['cluster_analysis']['overall_stats']['comment_count_max']=$comment_count_max;
	$_SESSION['cluster_analysis']['overall_stats']['comment_count_min']=$comment_count_min;
		
	//main cluster characteristics per each cluster (max/min type of things that distinguishes a cluster from others)
	
	//print stats into HTML
	$page_html_overall_stats .= '<tr>';
	$page_html_overall_stats .= '<td colspan="2">';
	$page_html_overall_stats .= '<fieldset class="analytics_metrics_time_subpanel">';
	$page_html_overall_stats .= '<legend style="font-weight: bold;">Clustering overview</legend>';
	
	$page_html_overall_stats .= "<div class='analytics_stats'><div class='analytics_stats_name'>Cluster count:</div> <div class='analytics_metric_value'>".$cluster_count ."</div>" . "</div>";
	
	$page_html_overall_stats .= "<div class='analytics_stats'><div class='analytics_stats_name'>Total idea count:</div> <div class='analytics_metric_value'>".$idea_count_total ."</div>" . "</div>";

	$page_html_overall_stats .= "<div class='analytics_stats'><div class='analytics_stats_name'>Last idea date:</div> <div class='analytics_metric_value'>".$last_idea_date."</div>" . "</div>";
	
	$page_html_overall_stats .= "<div class='analytics_tats'><div class='analytics_stats_name'>Ideas per Cluster</div> - <div class='analytics_metric_name'>avg: </div><div class='analytics_metric_value'>".$idea_count_avg."</div> <div class='analytics_metric_name'>max: </div><div class='analytics_metric_value'>".$idea_count_max. "</div> <div class='analytics_metric_name'>min: </div><div class='analytics_metric_value'>".$idea_count_min. "</div></div>";
	
	$page_html_overall_stats .= "<div class='analytics_tats'><div class='analytics_stats_name'>Comments per Cluster</div> - <div class='analytics_metric_name'>avg: </div><div class='analytics_metric_value'>".$comment_count_avg."</div> <div class='analytics_metric_name'>max: </div><div class='analytics_metric_value'>".$comment_count_max. "</div> <div class='analytics_metric_name'>min: </div><div class='analytics_metric_value'>".$comment_count_min. "</div></div>";
	
	$page_html_overall_stats .= '</fieldset>';
	$page_html_overall_stats .= '</td>';
	$page_html_overall_stats .= '</tr>';
	
	$page_html = $page_html_overall_stats; //. $page_html;
	
	
	/************ PRINT STATISTICS PER EACH CLUSTER **********/
	$current_cluster_number = 0;
	foreach ($ideas_clusters as $idea_cluster)
	{
		if ($current_cluster_number%2 == 0) $page_html .= '<tr>'; //for even cluster number start a new table row
		$page_html .= '<td class="analytics_metrics_subpanel">';
		$page_html .= '<div id="analytics_list_metrics">';
		$page_html .= '<div class="analytics-section-title"><div class="analytics-section-title-color-box" style="background-color: '.$d3_color_scale10[$current_cluster_number].';"></div>Cluster '.($current_cluster_number+1).'</div>';
		
		$tag_cluster = $tags_clusters[$current_cluster_number];
		$tag_cluster_count_total = count($tag_cluster); // includes tags that might have zero ideas
		$tag_cluster_count = count($idea_count_per_term_per_cluster[$current_cluster_number]); //only tags that have at least one idea in the cluster
		$page_html .= "<div class='analytics_stats'><div class='analytics_stats_name'>Theme</div> ($tag_cluster_count): <div class='analytics_metric_value' style='word-wrap: break-word;'>";
		
		$current_tag_count = 0;
		
		//save into session data
		$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['name']="CLUSTER ".($current_cluster_number+1);
		
		if ($idea_count_per_term_per_cluster[$current_cluster_number])
		foreach ($idea_count_per_term_per_cluster[$current_cluster_number] as $term_name => $idea_count) 
		{
			if ($current_tag_count==10) //show only 10 tags, hide the other and allow to expand on click
				$page_html .= '<span class="analytics_details_toggle_content">';
			if ($current_tag_count==0) {
				if (isset($term_ids_per_cluster[$current_cluster_number][$term_name])) 
					$page_html .= "<div class='analytics_important_term'>".$term_name." <div class='analytics_metric_value_bracket'>(".$idea_count.")</div></div>";
				else
					$page_html .= "<div class='analytics_unimportant_term'>".$term_name." <div class='analytics_metric_value_bracket'>(".$idea_count.")</div></div>";
			}
			else
			{
				if (isset($term_ids_per_cluster[$current_cluster_number][$term_name])) 
					$page_html .= ', '."<div class='analytics_important_term'>".$term_name." <div class='analytics_metric_value_bracket'>(".$idea_count.")</div></div>";
				else
					$page_html .= "<div class='analytics_unimportant_term'>".', '.$term_name." <div class='analytics_metric_value_bracket'>(".$idea_count.")</div></div>";
			}
			$current_tag_count++;

			//save to session data
			$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['term'][$term_name]=$idea_count;
		}
		
		if ($current_tag_count>10) $page_html .= "</span> <a class='analytics_details_toggle_link' href='#' >more</a>";
		$page_html .= "</div></div>";
		
		if ($idea_count_total!=0)
			$idea_count_percent = round($idea_count_per_cluster[$current_cluster_number] / $idea_count_total *100,2);
		else
			$idea_count_percent = 0; 
		$page_html .= "<div class='analytics_stats'><div class='analytics_stats_name'>Ideas Count:</div> <div class='analytics_metric_value'>".$idea_count_per_cluster[$current_cluster_number]."</div>";
		//add min/max insights with respect to other clusters
		if ($idea_count_per_cluster[$current_cluster_number]==$max_idea_count && $idea_count_per_cluster[$current_cluster_number]!=$min_idea_count) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_max_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>max</div>)</div></div>"; else
		if ($idea_count_per_cluster[$current_cluster_number]!=$max_idea_count && $idea_count_per_cluster[$current_cluster_number]==$min_idea_count) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_min_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>min</div>)</div></div>"; else
		if ($idea_count_per_cluster[$current_cluster_number]==$max_idea_count && $idea_count_per_cluster[$current_cluster_number]==$min_idea_count && $idea_count_percent==100) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_excl_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>exclusive</div>)</div></div>"; else
		if (count(array_unique($idea_count_per_cluster)) == 1 && count($idea_count_per_cluster) == $cluster_count) $page_html .= " <div class='analytics_metric_value_bracket'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>equilibrium</div>)</div></div>"; else
			$page_html .= " <div class='analytics_metric_value_bracket'>($idea_count_percent"."%)</div></div>";
		
		//comment count stat
		if ($comment_count_total!=0)
			$comment_count_percent = round($comment_count_per_cluster[$current_cluster_number] / $comment_count_total *100,2);
		else
			$comment_count_percent = 0; 
		$page_html .= "<div class='analytics_stats'><div class='analytics_stats_name'>Comment Count:</div> <div class='analytics_metric_value'>".$comment_count_per_cluster[$current_cluster_number]."</div>";
		//add min/max insights with respect to other clusters
		if ($comment_count_per_cluster[$current_cluster_number]==$max_comment_count && $comment_count_per_cluster[$current_cluster_number]!=$min_comment_count) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_max_insight'>($comment_count_percent"."% - <div class='analytics_metric_overall_insight'>max</div>)</div></div>"; else
		if ($comment_count_per_cluster[$current_cluster_number]!=$max_comment_count && $comment_count_per_cluster[$current_cluster_number]==$min_comment_count) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_min_insight'>($comment_count_percent"."% - <div class='analytics_metric_overall_insight'>min</div>)</div></div>"; else
		if ($comment_count_per_cluster[$current_cluster_number]==$max_comment_count && $comment_count_per_cluster[$current_cluster_number]==$min_comment_count && $comment_count_percent==100) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_excl_insight'>($comment_count_percent"."% - <div class='analytics_metric_overall_insight'>exclusive</div>)</div></div>"; else
		if (count(array_unique($comment_count_per_cluster)) == 1 && count($comment_count_per_cluster) == $cluster_count) $page_html .= " <div class='analytics_metric_value_bracket'>($comment_count_percent"."% - <div class='analytics_metric_overall_insight'>equilibrium</div>)</div></div>"; else
			$page_html .= " <div class='analytics_metric_value_bracket'>($comment_count_percent"."%)</div></div>";
		
		//comment per idea value
		$page_html .= "<div class='analytics_stats'><div class='analytics_stats_name'>Comment Per Idea Ratio:</div> <div class='analytics_metric_value'>".$comment_idea_ratio_per_cluster[$current_cluster_number]."</div>";
		//add min/max insights with respect to other clusters
		if ($comment_idea_ratio_per_cluster[$current_cluster_number]==$max_comment_idea_ratio && $comment_idea_ratio_per_cluster[$current_cluster_number]!=$min_comment_idea_ratio) $page_html .= " <div class='analytics_metric_overall_insight analytics_metric_value_max_insight'>(max)</div></div>"; else
		if ($comment_idea_ratio_per_cluster[$current_cluster_number]!=$max_comment_idea_ratio && $comment_idea_ratio_per_cluster[$current_cluster_number]==$min_comment_idea_ratio) $page_html .= " <div class='analytics_metric_overall_insight analytics_metric_value_min_insight'>(min)</div></div>"; else
		if (count(array_unique($comment_idea_ratio_per_cluster)) == 1 && count($comment_idea_ratio_per_cluster) == $cluster_count) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_overall_insight'>(equilibrium)</div></div>"; else
			$page_html .= "</div>";
			
		//save all data to session variable (for later CSV exports)
		$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['idea_count']=$idea_count_per_cluster[$current_cluster_number];
		$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['comment_count']=$comment_count_per_cluster[$current_cluster_number];
		$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['comment_per_idea']=$comment_idea_ratio_per_cluster[$current_cluster_number];
		
		//idea status stats
		$page_html .= "<div class='analytics_tats'><div class='analytics_stats_name' style='text-decoration:underline'>Idea Workflow breakdown:</div>";
		foreach ($workflow_idea_count_per_cluster as $workflow_type_name => $idea_count_in_clusters)
		{
			if (isset($idea_count_in_clusters[$current_cluster_number])) {
				if (strlen($workflow_type_name)>18)
					$state_title = substr($workflow_type_name,0,15)."...";
				else
					$state_title = $workflow_type_name;
				
				$page_html .= "<div class='analytics_tats'><div class='analytics_stats_name'>".$state_title."</div>: <div class='analytics_metric_value'>".$idea_count_in_clusters[$current_cluster_number]."</div>";
				//add min/max insights with respect to other clusters
				$idea_count_percent = round($idea_count_in_clusters[$current_cluster_number] / array_sum($idea_count_in_clusters) *100,2);
				if ($idea_count_in_clusters[$current_cluster_number]==$max_idea_count_per_workflow_state[$workflow_type_name] && $idea_count_in_clusters[$current_cluster_number]!=$min_idea_count_per_workflow_state[$workflow_type_name]) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_max_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>max</div>)</div></div>"; else
				if ($idea_count_in_clusters[$current_cluster_number]!=$max_idea_count_per_workflow_state[$workflow_type_name] && $idea_count_in_clusters[$current_cluster_number]==$min_idea_count_per_workflow_state[$workflow_type_name]) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_min_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>min</div>)</div></div>"; else
				if ($idea_count_in_clusters[$current_cluster_number]==$max_idea_count_per_workflow_state[$workflow_type_name] && $idea_count_in_clusters[$current_cluster_number]==$min_idea_count_per_workflow_state[$workflow_type_name] && $idea_count_percent==100) $page_html .= " <div class='analytics_metric_value_bracket analytics_metric_value_excl_insight'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>exclusive</div>)</div></div>"; else
				if (count(array_unique($idea_count_in_clusters)) == 1 && count($idea_count_in_clusters) == $cluster_count) $page_html .= " <div class='analytics_metric_value_bracket'>($idea_count_percent"."% - <div class='analytics_metric_overall_insight'>equilibrium</div>)</div></div>"; else
					$page_html .= " <div class='analytics_metric_value_bracket'>($idea_count_percent"."%)</div></div>";
			
				//save information into session for later CSV export
				$_SESSION['cluster_analysis']['cluster_data']["cluster_$current_cluster_number"]['workflow'][$state_title]=$idea_count_in_clusters[$current_cluster_number];
			}
		}
			
		$page_html .= '</div>';
		$page_html .= '</td>';
				
		if ($current_cluster_number%2 != 0 || ($current_cluster_number%2 == 0 && $current_cluster_number==($cluster_count-1))) //for odd cluster numbers close the table row (or for even when there is no clusters more to display)
			$page_html .= '</tr>';
		
		$current_cluster_number++;
	}
		
	$page_html .= '</table>';
	
	$page_html .= '<a href="?q=analytics_export/cluster_analytics/csv" style="float: right;"><img title="CSV Export" src="'.drupal_get_path('module', 'ideastream_analytics') .'/img/csv_icon.png"/></a>';
	
	return $page_html;
}

function theme_idea_analytics_cluster_analysis($form) {
	unset($form['cluster_statistics']['advanced_settings']['term_blacklist']['#maxlength']);
  return drupal_render($form);
}

function idea_analytics_cluster_analysis_submit($form, &$form_state) {
	//mark form rebuild to keep form_state variable 
	$form_state['rebuild'] = TRUE;
}

//form operations

/* 
 * Calculates clusters of ideas given a particular annotation type criteria
 * @type - annotation type that determines clustering criteria (tags,characteristics,categories, relations, frequent keywords)
 * @return table with ideas grouped in clusters and with coords in the clustering dimension space, additionally centroids that are centres of clusters
 * 			The function resutnr a number instead of array if there was an error. 
 *			Error codes: 
 *			-1 - no annotations with terms of selected type exist
 */
function idea_analytics_cluster_analysis_calculate($type = 'Characteristics',$settings = NULL) {

//temporarly for debugging
error_reporting(E_ALL);
	  
  	/* generate matrix with {ideas x annotations} ,where annotations are determined by $type = (tags, characteristics ...) */
  	if ($settings) $idea_characteristic_array = idea_analytics_cluster_analysis_generate_matrix($type,$settings['max_terms'],$settings['blacklist']);
  	else
  		$idea_characteristic_array = idea_analytics_cluster_analysis_generate_matrix($type);
  	$idea_characteristic_value_array = $idea_characteristic_array['VALUES'];
  	$idea_characteristic_row_info_array = $idea_characteristic_array['ROW_INFO']; //ideas
  	$idea_characteristic_column_info_array = $idea_characteristic_array['COLUMN_INFO']; //characteristics
  
  	//check if there are any terms and if the matrixes were constructed
  	if (count($idea_characteristic_value_array)==0 || count($idea_characteristic_row_info_array)==0 || count($idea_characteristic_column_info_array)==0)
  		return -1;
  	
  	/* calculate SVD (singular value decomposition) matrixes (this is done to have annotations and ideas in the same dimension space) */

  	$M = new Matrix($idea_characteristic_value_array);
    $E = new SingularValueDecomposition($M);
	$U_matrix = $E->getU(); //idea matrix
	$V_matrix = $E->getV(); //characteristics matrix (not transposed, each row is a characteristic)
	$S_matrix = $E->getS();
	
	/* extract two dimensions to cluster by particular dimensions from the matrixes generated by SVD (here using 2nd and 3rd dimension, numbering starts from 0) */
	$U_matrix_xy= idea_analytics_cluster_analysis_prepare_coordinate_matrixes($U_matrix->getArray(),1,2);
	$V_matrix_xy= idea_analytics_cluster_analysis_prepare_coordinate_matrixes($V_matrix->getArray(),1,2);
	
	/*modify the coordinate vectors by adding at the end an array with item id and item title (in order to be able to distinguish and name points after clustering) */
	$U_matrix_xy_mutated = array();
	foreach ($U_matrix_xy as $id => $coords) {
		$U_matrix_xy_mutated[$id] = $coords;
		$U_matrix_xy_mutated[$id][] = $idea_characteristic_row_info_array[$id];
	}
	
	$V_matrix_xy_mutated = array();
	foreach ($V_matrix_xy as $id => $coords) {
		$V_matrix_xy_mutated[$id] = $coords;
		$V_matrix_xy_mutated[$id][] = $idea_characteristic_column_info_array[$id];
	}
		
	/* run k-means clustering */

	//cluster characteristics together with ideas
	//$data = array_merge($U_matrix_xy_mutated, $V_matrix_xy_mutated);
	
	//cluster only ideas (leave out characteristics for later)
	$data = $U_matrix_xy_mutated;
	
	if ($settings['cluster_count'] && is_numeric($settings['cluster_count'])) $result = kMeans($data, $settings['cluster_count']);
	else
		$result = kMeans($data, 4);
		
	//add terms to clusters - use min distance from the cluster to assign terms (basically a single loop of kMeans clustering but without altering centroids)
	$result = assignClusterTerms($result,$V_matrix_xy_mutated);
	
	return $result;
}

//function generates a matrix
//@param type - type of matrix to be generated :
//	- Characteristics - characteristics count per idea
//	- Tags - tag count per idea
//	- Categories - category count per idea
//	- Frequent keywords - keyword count (from drupal search) per idea
//	- Relationships - relationship type count per idea
//@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
//@param blacklist - list of terms to be ignored when generating the matrix
//@return array[ROW_DATA, COLUMN_DATA,VALUE_MATRIX]
// CONCEPT_DATA - array[CONCEPT_NAME, CONCEPT_ID]
// IDEA_DATA - array[IDEA_NAME, IDEA_ID]
function idea_analytics_cluster_analysis_generate_matrix($type = 'Characteristics', $term_limit = 50, $blacklist = '') {
	$result = array();
	switch($type) {
		case 'Characteristics':
			$result = idea_analytics_cluster_analysis_characteristic_count($term_limit, $blacklist);
			break;
		case 'Tags':
			$result = idea_analytics_cluster_analysis_tag_count($term_limit, $blacklist);
			break;
		case 'Categories':
			$result = idea_analytics_cluster_analysis_category_count($term_limit, $blacklist);
			break;
		case 'Frequent keywords':
			$result = idea_analytics_cluster_analysis_keyword_count($term_limit, $blacklist);
			break;
		case 'Relationships':
			$result = idea_analytics_cluster_analysis_relationship_count($term_limit, $blacklist);
			break;
		default:		
			break;
	}
	
	ob_start();
	var_dump($result);
	$result_output = ob_get_clean();
	//dd($result_output);
	
	return $result;
}

/*************************************	SOURCE DATA GENERATION FUNCTIONS *****************************************/
/*****************************************************************************************************************/

/* Returns a metric with Number of Idea Tags by Tag name per each idea (rows are ideas, columns are tag names, values are 0 or 1)
	[ROW_INFO, COLUMN_INFO, VALUES] 
	ROW_INFO[idea_nid] = [idea_nid, idea_title]
	COLUMN_INFO[] = [term_id, term_name]
	VALUES[idea][tag] = 1 or 0 (exists / doesn't exist)
	@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
	@param blacklist - list of terms to be ignored when generating the matrix
	*/
function idea_analytics_cluster_analysis_tag_count($term_limit = 50, $blacklist = '') {
		
	//get 'Idea Tags' vocabulary id
	$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Tags'");
	if ($data = db_fetch_object($result)) {
	   $vocabulary_id=$data->vid;
	}
	
	if (isset($vocabulary_id)) { //if idea classification vocabulary exists
		//generate empty characteristic vector for top (by usage count) characteristics available
		$empty_characteristic_vector = array();
		$charactersitic_data = array();	
		$charactersitic_id_list = array();
		
		$blacklist_string = "'".implode("','", $blacklist)."'";
		
		$result = db_query("SELECT term_data.name AS term_name, term_data.tid AS term_id, COUNT(term_data.tid) AS term_count FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY term_count DESC LIMIT 0,$term_limit");
		while ($data = db_fetch_object($result)) {
			$empty_characteristic_vector[$data->term_id] = 0;
			
			$charactersitic_data[] = array('tag',$data->term_id,$data->term_name);
			$charactersitic_id_list[] = $data->term_id;
		}
		

		if (isset($charactersitic_id_list) && count($charactersitic_id_list)>0)
			$charactersitic_id_string = implode(",", $charactersitic_id_list);
		else
			$charactersitic_id_string = 'NULL'; 
		//get annotations and their count
  		$result = db_query("SELECT node.nid AS idea_id, node.title AS idea_title, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.tid IN ($charactersitic_id_string)");

		$characteristic_count = array();
		$idea_data = array();

		$ind = 0;
		$test_ind = 0;
		//$old = memory_get_usage();
		while ($data = db_fetch_object($result)) {

			if (!isset($characteristic_count[$data->idea_id])) { $characteristic_count[$data->idea_id] = $empty_characteristic_vector; $ind++;};
			$characteristic_count[$data->idea_id][$data->term_id]=1;
			if (!isset($idea_data[$data->idea_id])) $idea_data[$data->idea_id] = array('idea',$data->idea_id,$data->idea_title);
		}
		
		//$mem = memory_get_usage();

		//$size = abs($mem - $old);
		
		$characteristic_count = array_values($characteristic_count);
		foreach ($characteristic_count as $index => $characteristic_count_idea) 
			$characteristic_count[$index] = array_values($characteristic_count_idea);
		//change to regular index numbers (before rows were identified by idea nid)
		$idea_data = array_values($idea_data);
	}
	return array('ROW_INFO' => $idea_data, 'COLUMN_INFO' => $charactersitic_data, 'VALUES' => $characteristic_count);
}

/* Returns a metric with Number of Idea Categories by Category name per each idea (rows are ideas, columns are tag names, values are 0 or 1)
	[ROW_INFO, COLUMN_INFO, VALUES] 
	ROW_INFO[idea_nid] = [idea_nid, idea_title]
	COLUMN_INFO[] = [term_id, term_name]
	VALUES[idea][tag] = 1 or 0 (exists / doesn't exist)
	@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
	@param blacklist - list of terms to be ignored when generating the matrix
	*/
function idea_analytics_cluster_analysis_category_count($term_limit = 50, $blacklist = '') {
		
	//get 'Idea Tags' vocabulary id
	$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Categories'");
	if ($data = db_fetch_object($result)) {
	   $vocabulary_id=$data->vid;
	}
	
	if (isset($vocabulary_id)) { //if idea classification vocabulary exists
		//generate empty characteristic vector for all characteristics available
		$empty_characteristic_vector = array();
		$charactersitic_data = array();	
		$charactersitic_id_list = array();
		$blacklist_string = "'".implode("','", $blacklist)."'";
		
		$result = db_query("SELECT term_data.name AS term_name, term_data.tid AS term_id, COUNT(term_data.tid) AS term_count FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY term_count DESC LIMIT 0,$term_limit");
		while ($data = db_fetch_object($result)) {
			$empty_characteristic_vector[$data->term_id] = 0;

			$charactersitic_data[] = array('category',$data->term_id,$data->term_name);
			$charactersitic_id_list[] = $data->term_id;
		}
		
		//get annotations and their count
		if (isset($charactersitic_id_list) && count($charactersitic_id_list)>0)
			$charactersitic_id_string = implode(",", $charactersitic_id_list);
		else
			$charactersitic_id_string = 'NULL';
		
  		$result = db_query("SELECT node.nid AS idea_id, node.title AS idea_title, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.tid IN ($charactersitic_id_string)");

		$characteristic_count = array();
		$idea_data = array();

		$ind = 0;
		$test_ind = 0;
		//$old = memory_get_usage();
		while ($data = db_fetch_object($result)) {

			if (!isset($characteristic_count[$data->idea_id])) { $characteristic_count[$data->idea_id] = $empty_characteristic_vector; $ind++;};
			$characteristic_count[$data->idea_id][$data->term_id]=1;
			if (!isset($idea_data[$data->idea_id])) $idea_data[$data->idea_id] = array('idea',$data->idea_id,$data->idea_title);
		}
		
		//$mem = memory_get_usage();

		//$size = abs($mem - $old);
		
		$characteristic_count = array_values($characteristic_count);
		foreach ($characteristic_count as $index => $characteristic_count_idea) 
			$characteristic_count[$index] = array_values($characteristic_count_idea);
		//change to regular index numbers (before rows were identified by idea nid)
		$idea_data = array_values($idea_data);
	}
	return array('ROW_INFO' => $idea_data, 'COLUMN_INFO' => $charactersitic_data, 'VALUES' => $characteristic_count);
}


/* Returns a metric with Number of Idea Characteristics by type per each idea (rows are ideas, columns are characteristics)
	[ROW_INFO, COLUMN_INFO, VALUES] 
	ROW_INFO[idea_nid] = [idea_nid, idea_title]
	COLUMN_INFO[] = [term_id, term_name]
	VALUES[idea][characteristic] = 1 or 0 (exists / doesn't exist)
	@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
	@param blacklist - list of terms to be ignored when generating the matrix
	*/
function idea_analytics_cluster_analysis_characteristic_count($term_limit = 90, $blacklist = '') {
	$idea_characteristics = array();
		
	//get 'Idea Classification' vocabulary id
	$result = db_query("SELECT vocabulary.vid FROM vocabulary WHERE vocabulary.name = 'Idea Classification'");
	if ($data = db_fetch_object($result)) {
	   $vocabulary_id=$data->vid;
	}
	
	if (isset($vocabulary_id)) { //if idea classification vocabulary exists
		//generate empty characteristic vector for all characteristics available
		$empty_characteristic_vector = array();
		$charactersitic_data = array();	
		$charactersitic_id_list = array();

		$blacklist_string = "'".implode("','", $blacklist)."'";
		
		$result = db_query("SELECT term_data.name AS term_name, term_data.tid AS term_id, COUNT(term_data.tid) AS term_count FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.name NOT IN ($blacklist_string) GROUP BY term_id ORDER BY term_count DESC LIMIT 0,$term_limit");
		while ($data = db_fetch_object($result)) {
			$empty_characteristic_vector[$data->term_id] = 0;

			$charactersitic_data[] = array('characteristic',$data->term_id,$data->term_name);
			$charactersitic_id_list[] = $data->term_id;
		}
		
	
		//get annotations and their count		
		if (isset($charactersitic_id_list) && count($charactersitic_id_list)>0)
			$charactersitic_id_string = implode(",", $charactersitic_id_list);
		else
			$charactersitic_id_string = 'NULL';
		
  		$result = db_query("SELECT node.nid AS idea_id, node.title AS idea_title, term_data.name AS term_name, term_data.tid AS term_id FROM node JOIN term_node ON term_node.nid = node.nid JOIN term_data ON term_data.tid = term_node.tid WHERE node.type='idea' AND node.status=1 AND term_data.vid=$vocabulary_id AND term_data.tid IN ($charactersitic_id_string)");

		$characteristic_count = array();
		$idea_data = array();

		$ind = 0;
		$test_ind = 0;
		
		//$old = memory_get_usage();
		while ($data = db_fetch_object($result)) {
			if (!isset($characteristic_count[$data->idea_id])) { $characteristic_count[$data->idea_id] = $empty_characteristic_vector; $ind++;};
			$characteristic_count[$data->idea_id][$data->term_id]=1;
			if (!isset($idea_data[$data->idea_id])) $idea_data[$data->idea_id] = array('idea',$data->idea_id,$data->idea_title);
			$test_ind++;
		}
		//$mem = memory_get_usage();
		//$size = abs($mem - $old);
		
		$characteristic_count = array_values($characteristic_count);
		foreach ($characteristic_count as $index => $characteristic_count_idea) 
			$characteristic_count[$index] = array_values($characteristic_count_idea);
		//change to regular index numbers (before rows were identified by idea nid)
		$idea_data = array_values($idea_data);
	}
	return array('ROW_INFO' => $idea_data, 'COLUMN_INFO' => $charactersitic_data, 'VALUES' => $characteristic_count);
}

/* Returns a metric with Number of Idea Relationships by Relationship name per each idea (rows are ideas, columns are relationship names, values are 0 or 1)
	[ROW_INFO, COLUMN_INFO, VALUES] 
	ROW_INFO[idea_nid] = [idea_nid, idea_title]
	COLUMN_INFO[] = [term_id, term_name]
	VALUES[idea][tag] = 1 or 0 (exists / doesn't exist)
	@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
	@param blacklist - list of terms to be ignored when generating the matrix
	*/
function idea_analytics_cluster_analysis_relationship_count($term_limit = 50, $blacklist = '') {
	//$relationship_type_limit = 50;
			
	if (module_exists('similar_ideas')) { //if IdeaStream Similarity is installed calculate amount of relationships and related ideas
		//generate empty characteristic vector for top (by usage count) characteristics available
		$empty_characteristic_vector = array();
		$charactersitic_data = array();	
		$charactersitic_id_list = array();
		
		$blacklist_string = "'".implode("','", $blacklist)."'";
		$result = db_query("SELECT relationship_type AS term_name, COUNT(idea1) AS idea_count FROM similar_ideas_relations WHERE relationship_type NOT IN ($blacklist_string) GROUP BY relationship_type ORDER BY idea_count DESC LIMIT 0,$term_limit"); //full list of relationships that are being used in the system (each at least once), ordered by use count

		while ($data = db_fetch_object($result)) {
			$empty_characteristic_vector[$data->term_name] = 0;

			$charactersitic_data[] = array('Relationship Type',$data->term_name,$data->term_name);
			$charactersitic_id_list[] = $data->term_name;
		}
		
		//get annotations and their count
		if (isset($charactersitic_id_list) && count($charactersitic_id_list)>0)
			$charactersitic_id_string = "'".implode("','", $charactersitic_id_list)."'";
		else
			$charactersitic_id_string = 'NULL';
		
  		$result = db_query("SELECT idea1 AS idea_id, relationship_type AS term_name FROM similar_ideas_relations WHERE relationship_type IN ($charactersitic_id_string)");
		
		$characteristic_count = array();
		$idea_data = array();

		$ind = 0;
		$test_ind = 0;
		//$old = memory_get_usage();
		while ($data = db_fetch_object($result)) {

			if (!isset($characteristic_count[$data->idea_id])) { $characteristic_count[$data->idea_id] = $empty_characteristic_vector; $ind++;};
			$characteristic_count[$data->idea_id][$data->term_name]++; //=1;
			if (!isset($idea_data[$data->idea_id])) $idea_data[$data->idea_id] = array('idea',$data->idea_id,"Placeholder idea title");
		}
		
		//$mem = memory_get_usage();
		//$size = abs($mem - $old);
		
		$characteristic_count = array_values($characteristic_count);
		foreach ($characteristic_count as $index => $characteristic_count_idea) 
			$characteristic_count[$index] = array_values($characteristic_count_idea);
		//change to regular index numbers (before rows were identified by idea nid)
		$idea_data = array_values($idea_data);
	} 

	return array('ROW_INFO' => $idea_data, 'COLUMN_INFO' => $charactersitic_data, 'VALUES' => $characteristic_count);
}

/* Returns a metric with Number of Idea Keywords by Keyword name per each idea (rows are ideas, columns are keyword names, values are 0 or 1)
   The keywords are taken from Drupal search index based on Drupal estimation of most important words for the index (i.e. most frequent words that have the highest 		average score, where score is calcualted based on tags surrounding words in documents). Furthermore, the keywords are filtered against common stem words like 'and' 	'is' , 'are' etc..
	[ROW_INFO, COLUMN_INFO, VALUES] 
	ROW_INFO[idea_nid] = [idea_nid, idea_title]
	COLUMN_INFO[] = [term_id, term_name]
	VALUES[idea][tag] = 1 or 0 (exists / doesn't exist)
	@param term_limit - force limit of terms associated with ideas that are taken to generate the matrix (the top used terms are always taken first)
	@param blacklist - list of terms to be ignored when generating the matrix
	*/
function idea_analytics_cluster_analysis_keyword_count($term_limit = 50, $blacklist = '') {
	
	//list of common terms to be excluded from the keyword list
	$common_terms_list = "'the','and','for','that','not','you','add','idea','make','more','all','there','have','is','this','with','should','others','nothing','are','but','can','like','would','use','when','from','could','better','some','they','ideas','what','just','also','will','think','easy','which','need','1'";
	$keyword_limit = 50; //limit of keywords to be considered (the number is limited to speed up the LSA/cluster calculation)
		
	//generate empty characteristic vector for top (by usage count) characteristics available
	$empty_characteristic_vector = array();
	$charactersitic_data = array();
	$charactersitic_id_list = array();
	
	$blacklist_string = "'".implode("','", $blacklist)."'";
	$result = db_query("SELECT word AS term_name ,count FROM search_total WHERE word NOT IN ($blacklist_string) ORDER BY count ASC LIMIT 0,$term_limit");
	//$result = db_query("SELECT term_data.name AS term_name, term_data.tid AS term_id FROM term_data WHERE term_data.vid=$vocabulary_id");
	while ($data = db_fetch_object($result)) {
		$empty_characteristic_vector[$data->term_name] = 0;
		
		$charactersitic_data[] = array('Keyword',$data->term_name,$data->term_name);
		$charactersitic_id_list[] = $data->term_name;
	}
	
	//get annotations and their count
	if (isset($charactersitic_id_list) && count($charactersitic_id_list)>0)
		$charactersitic_id_string = "'".implode("','", $charactersitic_id_list)."'";
	else
		$charactersitic_id_string = 'NULL';
	
	$result = db_query("SELECT sid AS idea_id, word AS term_name FROM search_index WHERE word IN ($charactersitic_id_string)");

	$characteristic_count = array();
	$idea_data = array();

	$ind = 0;
	$test_ind = 0;
	//$old = memory_get_usage();
	while ($data = db_fetch_object($result)) {

		if (!isset($characteristic_count[$data->idea_id])) { $characteristic_count[$data->idea_id] = $empty_characteristic_vector; $ind++;};
		$characteristic_count[$data->idea_id][$data->term_name]=1;
		if (!isset($idea_data[$data->idea_id])) $idea_data[$data->idea_id] = array('idea',$data->idea_id,"Placeholder idea title");
	}
	
	//$mem = memory_get_usage();
	//$size = abs($mem - $old);
	
	$characteristic_count = array_values($characteristic_count);
	foreach ($characteristic_count as $index => $characteristic_count_idea) 
		$characteristic_count[$index] = array_values($characteristic_count_idea);
	//change to regular index numbers (before rows were identified by idea nid)
	$idea_data = array_values($idea_data);
		
	return array('ROW_INFO' => $idea_data, 'COLUMN_INFO' => $charactersitic_data, 'VALUES' => $characteristic_count);
}



/*******************************	END OF SOURCE DATA GENERATION FUNCTIONS **************************************/
/*****************************************************************************************************************/


/* 
 * Takes a two dimensional array as input and outs a new array with only two selected columns from the second dimension
 * (this function is used to select only two dimensions from the SVD matrixes in order to plot them to xy scatter chart)
 * @param matrix - matrix to process
 * @param dimension_one - first dimension index
 * @param dimension_two - second dimension index
 * @return matrix with values of from two selected dimensions
 */
function idea_analytics_cluster_analysis_prepare_coordinate_matrixes($matrix, $dimension_one, $dimension_two) {
	foreach ($matrix as $key => $value)
		$matrix[$key] = array($matrix[$key][$dimension_one], $matrix[$key][$dimension_two]);
	return $matrix;
}

/*
 * Renders a scatterplot in HTML markup using bluff charts based on passed coordinate values 
 * @param data - coordinates of points to be plotted into the scatter plot
 * @return HTML with the chart
 */
function idea_analytics_cluster_analysis_render_chart($data = array()) {
	//$comments_Month = idea_analytics_metric_ideas_per_month($dateMinusYear,$todayDate);
	
	$tooltip_labels_vector = array('1','2','3','4','5','6');
	$axis_labels_vector = array('1','2','3','4','5','6','7');
	$idea_count_vector = array(array(1,2),array(2,2),array(-1,2),array(1,-2),array(1,2),array(1,2));

	$axis_labels_vector[0]=	$tooltip_labels_vector[0];
	$axis_labels_vector[count($axis_labels_vector)-1]= $tooltip_labels_vector[count($axis_labels_vector)-1];
	
	$canvas = charts_graphs_get_graph('bluff');
	
	$canvas->title = "New Ideas Last Year (per Month)";
	$canvas->type = "dot"; // a chart type supported by the charting engine. See further in the doc for the list.
	$canvas->width = "600";
	
	$canvas->y_legend = "Y Legend";
	$canvas->colour = '#D54C78';
	$canvas->theme = 'pastel';   //keynote,37signals, rails_keynote, odeo, pastel, greyscale
	$canvas->series = array(
	'Ideas Amount' => $idea_count_vector,
	);

	$canvas->x_labels = $axis_labels_vector;
	$canvas->tooltip_labels = $tooltip_labels_vector;
	
	$page_html .= $canvas->get_chart();
	return $page_html;
}

/* helpers */

function arrayToCsv( array $fields, $delimiter = ';' ) {
    $delimiter_esc = preg_quote($delimiter, '/');
    $enclosure_esc = preg_quote($enclosure, '/');

    $csv_output = "";
    $column_length = count($fields);
    for ( $j=0; $j< $column_length; $j++) {
    	$test_i=0;
    	$field = $fields[$j];
    	if (is_array($field)) {
    		$row_length = count($field);
    		for ($i=0;$i<$row_length;$i++) 
    		{
    			if ($test_i==0)	$csv_output .= $field[$i]; else 
    				$csv_output .= $delimiter . $field[$i];
    			$test_i++;
    		}
    	}
    	$csv_output .= "\n";
    }

    return $csv_output;
}

function rowInfoArrayToCsv( array $fields, $delimiter = ';' ) {
    $delimiter_esc = preg_quote($delimiter, '/');
    $enclosure_esc = preg_quote($enclosure, '/');

    $csv_output = "";
    foreach ( $fields as $field ) {
    	$test_i=0;
    	if (is_array($field))
    		foreach ($field as $element) 
    		{
    			if ($test_i==0)	$csv_output .= $element; else 
    				$csv_output .= $delimiter . $element;
    			$test_i++;
    		}
    	$csv_output .= "\n";
    }

    return $csv_output;
}

/**************************** CLUSTERING k-means IMPLEMENTATION ***********************/

function kMeans($data, $k) {
	$centroids = initialiseCentroids($data, $k);
	/*
	ob_start();
	var_dump($centroids);
	$result = ob_get_clean();
	dd("Centroids: ".$result); */
	
	$mapping = array();

	$i = 0;
	while(true) {
		$new_mapping = assignCentroids($data, $centroids, $mapping);
		$changed = false;
		foreach($new_mapping as $documentID => $centroidID) {
			if(!isset($mapping[$documentID]) || $centroidID != $mapping[$documentID]) {
				$mapping = $new_mapping;
				$changed = true;
				break;
			}
		}
		if(!$changed || $i>99){ //max 99 loops
		//drupal_set_message("k-means loops:"+$i);
			return formatResults($mapping, $data, $centroids); 
		}
		$centroids  = updateCentroids($mapping, $data, $k); 
		$i++;
	}
}

function initialiseCentroids(array $data, $k) {
	$dimensions = count($data[0])-1;
	$centroids = array();
	$dimmax = array();
	$dimmin = array(); 
	foreach($data as $document) {
		foreach($document as $dim => $val) {
			if (!is_array($val)) //skip the final property vector that contains information not used for clustering (ids, names etc.)
			{
				if(!isset($dimmax[$dim]) || $val > $dimmax[$dim]) {
					$dimmax[$dim] = $val;
				}
				if(!isset($dimmin[$dim]) || $val < $dimmin[$dim]) {
					$dimmin[$dim] = $val;
				}
			}
		}
	}
	
	//make sure that $data its not an associative array and has integers as keys (because centroid init alt is based on random index selection)
	//$data = array_values($data);
	
	for($i = 0; $i < $k; $i++) {
		$centroids[$i] = initialiseCentroid($dimensions, $dimmax, $dimmin, $k, $data);
	}
	return $centroids;
}

function initialiseCentroid($dimensions, $dimmax, $dimmin, $k, $data) {
	
	//$k = 3;
	//$i++;
	$total = 0;
	$centroid = array();
	/*
	for($j = 0; $j < $dimensions; $j++) {
		$centroid[$j] = (rand($dimmin[$j] * 1000, $dimmax[$j] * 1000));
		//$centroid[$j] = $dimmin[$j] + ceil(abs($dimmax[$j] - $dimmin[$j]) / $i)  * $i;
		$total += $centroid[$j] * $centroid[$j];
	
	$centroid = normaliseValue($centroid, sqrt($total));
	*/
	$dataset_size = count($data);
	$observation_number = rand(0,$dataset_size-1);  //$k-1
	$centroid = $data[$observation_number];
	/*
	for($j = 0; $j < $dimensions; $j++) {
	} */
	return $centroid;
}

function formatResults($mapping, $data, $centroids) {
	$result  = array();
	$result['centroids'] = $centroids;
	foreach($mapping as $documentID => $centroidID) {
		$result[$centroidID][] = $data[$documentID]; //implode(',', $data[$documentID]);
	}
	return $result;
}

function assignCentroids($data, $centroids, $old_mapping = null, $old_clusters_size = null) {
	$mapping = array();
	$clusters_size = array();	
	
	foreach($data as $documentID => $document) {
		$minDist = 100;
		$minCentroid = null;
		foreach($centroids as $centroidID => $centroid) {
			$dist = 0;
			foreach($centroid as $dim => $value) {
				if (!is_array($value) && !is_array($document[$dim])) //ommit vectors that hold properties (i.e. name, ids etc.) - those values are stored in array as opposed to integers/floats
				{
					$dist += abs($value - $document[$dim]);
				}
			}
			if($dist < $minDist) {
				$minDist = $dist;
				$minCentroid = $centroidID;
			}
			/*
			if ($dist == $minDist) {
				$old_centroid_id = $old_mapping[$documentID];
				if ($old_clusters_size!=null && $old_clusters_size[$old_centroid_id] < $old_clusters_size[$centroidID]) //if the new group has more items then the old
				{
					$minDist = $dist;
					$minCentroid = $centroidID;
				}
			}*/
		}
		$mapping[$documentID] = $minCentroid;
		//$clusters_size[$minCentroid]++;
	}

	return $mapping;
}

function updateCentroids($mapping, $data, $k) {
	$centroids = array();
	$counts = array_count_values($mapping);

	foreach($mapping as $documentID => $centroidID) {
		foreach($data[$documentID] as $dim => $value) {
			if (!is_array($value))
				$centroids[$centroidID][$dim] += ($value/$counts[$centroidID]); 
		}
	}

	if(count($centroids) < $k) {
		$centroids = array_merge($centroids, initialiseCentroids($data, $k - count($centroids)));
	}

	return $centroids;
}

function normaliseValue(array $vector, $total) {
	foreach($vector as &$value) {
		$value = $value/$total;
	}
	return $vector;
}

/****************************** assign terms on top of clusters ***********************/

function assignClusterTerms($clusters,$termData) {

	foreach($termData as $documentID => $document) {
		$minDist = 100;
		$minCentroid = null;
		foreach($clusters['centroids'] as $centroidID => $centroid) {
			$dist = 0;
			foreach($centroid as $dim => $value) {
				if (!is_array($value) && !is_array($document[$dim])) //ommit vectors that hold properties (i.e. name, ids etc.) - those values are stored in array as opposed to integers/floats
				{
					$dist += abs($value - $document[$dim]);
				}
			}
			if($dist < $minDist) {
				$minDist = $dist;
				$minCentroid = $centroidID;
			}
		}
		$clusters[$minCentroid][] = $document; //add term to cluster
		//$clusters_size[$minCentroid]++;
	}

	return $clusters;
}
 
?>